import numpy as np

from typing import Optional, Dict

from mlserver.types import InferenceRequest, InferenceResponse
from mlserver.codecs import (
    NumpyCodec,
    RequestCodec,
    register_request_codec,
    get_decoded_or_raw,
)
from mlserver.codecs.lists import is_list_of

from typing import Any


TensorDict = Dict[str, np.ndarray]


@register_request_codec
class TensorDictCodec(RequestCodec):
    """
    The TensorDictCodec knows how to encode / decode a dictionary of tensors.
    """

    ContentType = "dict"

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        if not isinstance(payload, dict):
            return False

        return is_list_of(list(payload.values()), np.ndarray)

    @classmethod
    def encode_response(
        cls,
        model_name: str,
        payload: TensorDict,
        model_version: Optional[str] = None,
        **kwargs
    ) -> InferenceResponse:
        outputs = [
            NumpyCodec.encode_output(name, value, **kwargs)
            for name, value in payload.items()
        ]

        return InferenceResponse(
            model_name=model_name, model_version=model_version, outputs=outputs
        )

    @classmethod
    def decode_response(cls, response: InferenceResponse) -> TensorDict:
        return {
            response_output.name: NumpyCodec.decode_output(response_output)
            for response_output in response.outputs
        }

    @classmethod
    def encode_request(cls, payload: TensorDict, **kwargs) -> InferenceRequest:
        inputs = [
            NumpyCodec.encode_input(name, value, **kwargs)
            for name, value in payload.items()
        ]

        return InferenceRequest(inputs=inputs)

    @classmethod
    def decode_request(cls, request: InferenceRequest) -> TensorDict:
        return {
            request_input.name: get_decoded_or_raw(request_input)
            for request_input in request.inputs
        }



from typing import Union, Tuple, List, Optional

from mlflow.types.schema import Schema, ColSpec, TensorSpec, DataType

from mlserver.types import MetadataTensor, Parameters
from mlserver.codecs import (
    PandasCodec,
    NumpyCodec,
    StringCodec,
    Base64Codec,
    DatetimeCodec,
)
from mlserver.codecs.numpy import to_datatype

from .codecs import TensorDictCodec

InputSpec = Union[ColSpec, TensorSpec]

DefaultInputPrefix = "input-"
DefaultOutputPrefix = "output-"

_MLflowToContentType = {
    DataType.boolean: ("BOOL", NumpyCodec.ContentType),
    DataType.integer: ("INT32", NumpyCodec.ContentType),
    DataType.long: ("INT64", NumpyCodec.ContentType),
    DataType.float: ("FP32", NumpyCodec.ContentType),
    DataType.double: ("FP64", NumpyCodec.ContentType),
    DataType.string: ("BYTES", StringCodec.ContentType),
    DataType.binary: ("BYTES", Base64Codec.ContentType),
    DataType.datetime: ("BYTES", DatetimeCodec.ContentType),
}


def _get_content_type(input_spec: InputSpec) -> Tuple[str, str]:
    if isinstance(input_spec, TensorSpec):
        datatype = to_datatype(input_spec.type)
        content_type = NumpyCodec.ContentType
        return datatype, content_type

    # TODO: Check if new type, which may not exist
    return _MLflowToContentType[input_spec.type]


def _get_shape(input_spec: InputSpec) -> List[int]:
    if isinstance(input_spec, TensorSpec):
        return list(input_spec.shape)

    return [-1]


def to_metadata_tensors(
    schema: Schema, prefix=DefaultInputPrefix
) -> List[MetadataTensor]:
    metadata_tensors = []

    for idx, input_spec in enumerate(schema.inputs):
        datatype, content_type = _get_content_type(input_spec)
        shape = _get_shape(input_spec)

        name = input_spec.name if input_spec.name else f"{prefix}{idx}"

        metadata_tensors.append(
            MetadataTensor(
                name=name,
                datatype=datatype,
                shape=shape,
                parameters=Parameters(content_type=content_type),
            )
        )

    return metadata_tensors


def to_model_content_type(schema: Schema) -> Optional[str]:
    # This logic is based on MLflow's `mlflow.pyfunc._enforce_schema` method:
    # https://github.com/mlflow/mlflow/blob/ded7e447c20d259030260f1579693f9c5337a3ae/mlflow/pyfunc/__init__.py#L499
    if schema.is_tensor_spec():
        if schema.has_input_names():
            return TensorDictCodec.ContentType

        return NumpyCodec.ContentType

    return PandasCodec.ContentType
